import unittest
from defcon.tools.notifications import NotificationCenter
from defcon.test.testTools import NotificationTestObserver


class _TestObservable(object):

    def __init__(self, center, name):
        self.center = center
        self.name = name

    def postNotification(self, name):
        self.center.postNotification(name, self)


class NotificationCenterTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_addObserver_notification_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        self.assertTrue(center.hasObserver(observer, "A", observable))
        self.assertFalse(center.hasObserver(observer, "B", observable))

    def test_addObserver_notification_no_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", None)
        self.assertTrue(center.hasObserver(observer, "A", None))
        self.assertFalse(center.hasObserver(observer, "A", observable))

    def test_addObserver_no_notification_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, observable)
        self.assertTrue(center.hasObserver(observer, None, observable))
        self.assertFalse(center.hasObserver(observer, "A", observable))

    def test_addObserver_no_notification_no_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, None)
        self.assertTrue(center.hasObserver(observer, None, None))
        self.assertFalse(center.hasObserver(observer, "A", observable))
        self.assertFalse(center.hasObserver(observer, None, observable))
        self.assertFalse(center.hasObserver(observer, "A", None))

    def test_removeObserver(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.removeObserver(observer, "A", observable)
        self.assertFalse(center.hasObserver(observer, "A", observable))

    def test_removeObserver_notification_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.removeObserver(observer, "A", observable)
        self.assertFalse(center.hasObserver(observer, "A", observable))

    def test_removeObserver_notification_no_observable(self):
        center = NotificationCenter()
        _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", None)
        center.removeObserver(observer, "A", None)
        self.assertFalse(center.hasObserver(observer, "A", None))

    def test_removeObserver_no_notification_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, observable)
        center.removeObserver(observer, None, observable)
        self.assertFalse(center.hasObserver(observer, None, observable))

    def test_removeObserver_no_notification_no_observable(self):
        center = NotificationCenter()
        _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, None)
        center.removeObserver(observer, None, None)
        self.assertFalse(center.hasObserver(observer, None, None))

    def test_postNotification_notification_observable(self):
        # notification, observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.postNotification("A", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("A", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("B", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("B", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_postNotification_notification_no_observable(self):
        # notification, no observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", None)
        center.postNotification("A", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("A", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable2"))
        center.postNotification("B", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable2"))
        center.postNotification("B", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable2"))


    def test_postNotification_no_notification_observable(self):
        # no notification, observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, observable1)
        center.postNotification("A", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("A", observable2)
        center.postNotification("B", observable1)
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        center.postNotification("B", observable2)


    def test_postNotification_no_notification_no_observable(self):
        # no notification, no observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, None)
        center.postNotification("A", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("A", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable2"))
        center.postNotification("B", observable1)
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        center.postNotification("B", observable2)
        self.assertEqual(observer.stack[-1], ("B", "Observable2"))

    def test_hold_and_releaseHeldNotifications_all_notifications(self):
        "Hold all notifications"
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)

        center.holdNotifications()
        observable1.postNotification("A")
        self.assertEqual(len(observer.stack), 0)
        observable1.postNotification("A")
        self.assertEqual(len(observer.stack), 0)
        observable1.postNotification("B")
        self.assertEqual(len(observer.stack), 0)
        observable2.postNotification("C")
        self.assertEqual(len(observer.stack), 0)
        center.releaseHeldNotifications()
        self.assertEqual(observer.stack[-3], ("A", "Observable1"))
        self.assertEqual(observer.stack[-2], ("B", "Observable1"))
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        self.assertEqual(len(observer.stack), 3)

    def test_hold_and_releaseHeldNotifications_notifications_of_observable(self):
        "Hold all notifications of a specific observable"
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)

        center.holdNotifications(observable=observable1)
        observable1.postNotification("A")
        observable1.postNotification("A")
        observable1.postNotification("B")
        observable2.postNotification("C")
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        center.releaseHeldNotifications(observable=observable1)
        self.assertEqual(observer.stack[-2], ("A", "Observable1"))
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        self.assertEqual(len(observer.stack), 3)

    def test_hold_and_releaseHeldNotifications_notifications_of_observable(
            self):
        "Hold all notifications of a specific notification"
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)

        center.holdNotifications(notification="A")
        observable1.postNotification("A")
        observable1.postNotification("A")
        observable1.postNotification("B")
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        observable2.postNotification("C")
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        center.releaseHeldNotifications(notification="A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        self.assertEqual(len(observer.stack), 3)

    def test_areNotificationsHeld_all_held(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.holdNotifications()
        self.assertTrue(center.areNotificationsHeld())
        center.releaseHeldNotifications()
        self.assertFalse(center.areNotificationsHeld())

    def test_areNotificationsHeld_observable_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable2)
        center.holdNotifications(observable=observable1)
        self.assertTrue(center.areNotificationsHeld(observable=observable1))
        self.assertFalse(center.areNotificationsHeld(observable=observable2))
        center.releaseHeldNotifications(observable=observable1)
        self.assertFalse(center.areNotificationsHeld(observable=observable1))

    def test_areNotificationsHeld_notification_off(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.addObserver(observer, "notificationCallback", "B", observable)
        center.holdNotifications(notification="A")
        self.assertTrue(center.areNotificationsHeld(notification="A"))
        self.assertFalse(center.areNotificationsHeld(notification="B"))
        center.releaseHeldNotifications(notification="A")
        self.assertFalse(center.areNotificationsHeld(notification="A"))

    def test_areNotificationsHeld_observer_off(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer1 = NotificationTestObserver()
        observer2 = NotificationTestObserver()
        center.addObserver(observer1, "notificationCallback", "A", observable)
        center.addObserver(observer2, "notificationCallback", "A", observable)
        center.holdNotifications(observer=observer1)
        self.assertTrue(center.areNotificationsHeld(observer=observer1))
        self.assertFalse(center.areNotificationsHeld(observer=observer2))
        center.releaseHeldNotifications(observer=observer1)
        self.assertFalse(center.areNotificationsHeld(observer=observer1))

    def test_disable_enableNotifications_all_notifications(self):
        # disable all notifications
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.disableNotifications()
        observable1.postNotification("A")
        observable1.postNotification("B")
        observable2.postNotification("C")
        center.enableNotifications()
        observable1.postNotification("A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_disable_enableNotifications_specific_observable(self):
        # disable all notifications for a specific observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.disableNotifications(observable=observable1)
        observable1.postNotification("A")
        observable1.postNotification("B")
        observable2.postNotification("C")
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        center.enableNotifications(observable=observable1)
        observable1.postNotification("A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_disable_enableNotifications_specific_notification(self):
        # disable all notifications for a specific notification
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.disableNotifications(notification="A")
        observable1.postNotification("A")
        observable1.postNotification("B")
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        observable2.postNotification("C")
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        center.enableNotifications(notification="A")
        observable1.postNotification("A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_disable_enableNotifications_specific_observer(self):
        # disable all notifications for a specific observer
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.disableNotifications(observer=observer)
        observable1.postNotification("A")
        observable1.postNotification("B")
        observable2.postNotification("C")
        center.enableNotifications(observer=observer)
        observable1.postNotification("A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_areNotificationsDisabled_all_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable2)
        center.disableNotifications()
        self.assertTrue(center.areNotificationsDisabled())
        center.enableNotifications()
        self.assertFalse(center.areNotificationsDisabled())

    def test_areNotificationsDisabled_observable_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable2)
        center.disableNotifications(observable=observable1)
        self.assertTrue(
            center.areNotificationsDisabled(observable=observable1))
        self.assertFalse(
            center.areNotificationsDisabled(observable=observable2))
        center.enableNotifications(observable=observable1)
        self.assertFalse(
            center.areNotificationsDisabled(observable=observable1))

    def test_areNotificationsDisabled_notification_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable2)
        center.disableNotifications(notification="A")
        self.assertTrue(center.areNotificationsDisabled(notification="A"))
        self.assertFalse(center.areNotificationsDisabled(notification="B"))
        center.enableNotifications(notification="A")
        self.assertFalse(center.areNotificationsDisabled(notification="A"))

    def test_areNotificationsDisabled_observer_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observer1 = NotificationTestObserver()
        observer2 = NotificationTestObserver()
        center.addObserver(observer1, "notificationCallback", "A", observable1)
        center.addObserver(observer2, "notificationCallback", "A", observable1)
        center.disableNotifications(observer=observer1)
        self.assertTrue(center.areNotificationsDisabled(observer=observer1))
        self.assertFalse(center.areNotificationsDisabled(observer=observer2))
        center.enableNotifications(observer=observer1)
        self.assertFalse(center.areNotificationsDisabled(observer=observer1))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from setuptools import setup, find_packages, Command
import sys
from distutils import log


class bump_version(Command):

    description = "increment the package version and commit the changes"

    user_options = [
        ("major", None, "bump the first digit, for incompatible API changes"),
        ("minor", None, "bump the second digit, for new backward-compatible features"),
        ("patch", None, "bump the third digit, for bug fixes (default)"),
    ]

    def initialize_options(self):
        self.minor = False
        self.major = False
        self.patch = False

    def finalize_options(self):
        part = None
        for attr in ("major", "minor", "patch"):
            if getattr(self, attr, False):
                if part is None:
                    part = attr
                else:
                    from distutils.errors import DistutilsOptionError
                    raise DistutilsOptionError(
                        "version part options are mutually exclusive")
        self.part = part or "patch"

    def bumpversion(self, part, **kwargs):
        """ Run bump2version.main() with the specified arguments.
        """
        import bumpversion

        args = ['--verbose'] if self.verbose > 1 else []
        for k, v in kwargs.items():
            k = "--{}".format(k.replace("_", "-"))
            is_bool = isinstance(v, bool) and v is True
            args.extend([k] if is_bool else [k, str(v)])
        args.append(part)

        log.debug(
            "$ bumpversion %s" % " ".join(a.replace(" ", "\\ ") for a in args))

        bumpversion.main(args)

    def run(self):
        log.info("bumping '%s' version" % self.part)
        self.bumpversion(self.part)


class release(bump_version):
    """Drop the developmental release '.devN' suffix from the package version,
    open the default text $EDITOR to write release notes, commit the changes
    and generate a git tag.
    Release notes can also be set with the -m/--message option, or by reading
    from standard input.
    """

    description = "tag a new release"

    user_options = [
        ("message=", 'm', "message containing the release notes"),
    ]

    def initialize_options(self):
        self.message = None

    def finalize_options(self):
        import re

        current_version = self.distribution.metadata.get_version()
        if not re.search(r"\.dev[0-9]+", current_version):
            from distutils.errors import DistutilsSetupError
            raise DistutilsSetupError(
                "current version (%s) has no '.devN' suffix.\n       "
                "Run 'setup.py bump_version' with any of "
                "--major, --minor, --patch options" % current_version)

        message = self.message
        if message is None:
            if sys.stdin.isatty():
                # stdin is interactive, use editor to write release notes
                message = self.edit_release_notes()
            else:
                # read release notes from stdin pipe
                message = sys.stdin.read()

        if not message.strip():
            from distutils.errors import DistutilsSetupError
            raise DistutilsSetupError("release notes message is empty")

        self.message = "v{new_version}\n\n%s" % message

    @staticmethod
    def edit_release_notes():
        """Use the default text $EDITOR to write release notes.
        If $EDITOR is not set, use 'nano'."""
        from tempfile import mkstemp
        import os
        import shlex
        import subprocess

        text_editor = shlex.split(os.environ.get('EDITOR', 'nano'))

        fd, tmp = mkstemp(prefix='bumpversion-')
        try:
            os.close(fd)
            with open(tmp, 'w') as f:
                f.write("\n\n# Write release notes.\n"
                        "# Lines starting with '#' will be ignored.")
            subprocess.check_call(text_editor + [tmp])
            with open(tmp, 'r') as f:
                changes = "".join(
                    l for l in f.readlines() if not l.startswith('#'))
        finally:
            os.remove(tmp)
        return changes

    def run(self):
        log.info("stripping developmental release suffix")
        # drop '.dev0' suffix, commit with given message and create git tag
        self.bumpversion("release",
                         tag=True,
                         message="Release {new_version}",
                         tag_message=self.message)


needs_sphinx = {'build_sphinx', 'dist'}.intersection(sys.argv)
sphinx = ['sphinx'] if needs_sphinx else []
needs_pytest = {'pytest', 'test'}.intersection(sys.argv)
pytest_runner = ['pytest_runner'] if needs_pytest else []
needs_wheel = {'bdist_wheel'}.intersection(sys.argv)
wheel = ['wheel'] if needs_wheel else []
needs_bump2version = {'release', 'bump_version'}.intersection(sys.argv)
bump2version = ['bump2version'] if needs_bump2version else []

with open('README.rst', 'r') as f:
    long_description = f.read()

setup_params = dict(
    name="defcon",
    version="0.7.2",
    description="A set of flexible objects for representing UFO data.",
    long_description=long_description,
    author="Tal Leming",
    author_email="tal@typesupply.com",
    url="https://github.com/robotools/defcon",
    license="MIT",
    package_dir={"": "Lib"},
    packages=find_packages("Lib"),
    include_package_data=True,
    python_requires=">=3.6",
    setup_requires=pytest_runner + sphinx + wheel + bump2version,
    tests_require=[
        'pytest>=3.0.3',
    ],
    install_requires=[
        "fonttools[ufo,unicode] >= 4.10.0",
    ],
    extras_require={
        'pens': ["fontPens>=0.1.0"],
        'lxml': ["fonttools[lxml] >= 4.10.0"],
    },
    cmdclass={
        "release": release,
        "bump_version": bump_version,
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Multimedia :: Graphics :: Editors :: Vector-Based',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)


if __name__ == "__main__":
    setup(**setup_params)
